/* memory.c */

/* VIS-5D version 3.0 */

/*
vis5d program for visualizing five dimensional gridded data sets
Copyright (C) 1990, 1991, 1992  Bill Hibbard, Brian Paul, Dave Santek,
and Andre Battaiola.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



#include <assert.h>
#ifdef stellar
#  include <malloc.h>
#  include <memory.h>
#else
#  include <stdlib.h>
#  include <string.h>
#endif
#include <stdio.h>
#include "globals.h"
#include "sync.h"



struct mem {
   int        size;
   struct mem *prev;
   struct mem *next;
   short int        free, magic;
};

static struct mem *head, *tail;

static int memory_limit;
static int memory_used = 0;

static LOCK memlock;
static LOCK lrulock;


#define MEMSIZ sizeof(struct mem)
#define MAGIC 0x1234


static void *alloc(), dealloc();




/*** init_memory ******************************************************
   Initialize the memory management module.
   Input:  bytes - limit of memory to use.
**********************************************************************/
int init_memory( bytes )
int bytes;
{
   struct mem *m;

   assert( bytes==0 || bytes>=1024*1024 );

   memory_limit = bytes;

   if (memory_limit) {
      m = (struct mem *) malloc( memory_limit );
      assert(m);

      /*memset( m, 0, memory_limit ); */  /* clear and bring into RAM */

      m->size = bytes - sizeof(struct mem);
      m->prev = NULL;
      m->next = NULL;
      m->free = 1;
      m->magic = MAGIC;

      head = tail = m;
      memory_used = MEMSIZ;
   }
   else {
      memory_used = 0;
   }

   ALLOC_LOCK( memlock );
   ALLOC_LOCK( lrulock );
}



/*** allocate *********************************************************
   Allocate a block of memory.
   If there is not enough memory to satisfy the request, the least
   recently used graphics will be deallocated.
   Return:  address of memory block or NULL if out of memory.
**********************************************************************/
void *allocate( bytes )
int bytes;
{
   if (memory_limit==0) {
      /* just malloc */
      return (void *) malloc( bytes );
   }
   else {
      void *addr;
      int ma, d;

      do {
	 LOCK_ON( memlock );
	 addr = alloc( bytes, 0 );
	 LOCK_OFF( memlock );
	 if (addr) {
	    /* all done, return */
	    return addr;
	 }
	 /* We didn't find a free block large enough, */
	 /* try deallocating some graphics */
	 ma = mem_available();
	 LOCK_ON( lrulock );
	 if (ma==mem_available()) {
	    d = deallocate_lru();
	 }
	 LOCK_OFF( lrulock );
      } while (d>0);
      /* Couldn't deallocate anything, we're REALLY out of memory */
      return NULL;
   }
}



/*** pallocate ********************************************************
   Permanent allocate.  Same as allocate, above, but used to allocate
   memory which will NEVER be deallocated.
**********************************************************************/
void *pallocate( bytes )
int bytes;
{
   if (memory_limit==0) {
      /* just malloc */
      return (void *) malloc( bytes );
   }
   else {
      void *addr;
      int ma, d;

      do {
	 LOCK_ON( memlock );
	 addr = alloc( bytes, 1 );
	 LOCK_OFF( memlock );
	 if (addr) {
	    /* all done, return */
	    return addr;
	 }
	 /* We didn't find a free block large enough, */
	 /* try deallocating some graphics */
	 ma = mem_available();
	 LOCK_ON( lrulock );
	 if (ma==mem_available()) {
	    d = deallocate_lru();
	 }
	 LOCK_OFF( lrulock );
      } while (d>0);
      /* Couldn't deallocate anything, we're REALLY out of memory */
      return NULL;
   }
}




/*** deallocate *******************************************************
   Deallocate a block of memory.
   Input:  addr - address of block (if NULL, nothing happens)
           bytes - size of block (if <= zero, bytes is ignored)
**********************************************************************/
void deallocate( addr, bytes )
void *addr;
int bytes;
{
   LOCK_ON( memlock );
   if (addr) {
      if (memory_limit==0) {
	 free( addr );
      }
      else {
	 dealloc( addr, bytes );
      }
   }
   LOCK_OFF( memlock );
}



/*** mem_available ****************************************************
   Return the amount of available memory.
**********************************************************************/
int mem_available()
{
   if (memory_limit==0)
      return 1024*1024*1024;  /* a Gig ought to be enough */
   else
      return memory_limit-memory_used;
}



/*** mem_used ********************************************************
   Return the amount of memory used.
**********************************************************************/
int mem_used()
{
   if (memory_limit==0)
      return -1;
   else
     return memory_used;
}




/********************************************/
/****   Private functions                 ***/
/********************************************/


/* remove a node from linked list of memory nodes. */

static void remove_mem( pos )
struct mem *pos;
{
   /* make predecessor point to successor */
   if (pos->prev) {
      pos->prev->next = pos->next;
   }
   else {
      head = pos->next;
   }

   /* make successor point to predecessor */
   if (pos->next) {
      pos->next->prev = pos->prev;
   }
   else {
      tail = pos->prev;
   }

   pos->next = pos->prev = NULL;
}


/* used as a first guess to finding a free block of memory for */
/* allocation before doing a first-fit: */
static struct mem *guess = NULL;


static void *alloc( b, pflag )
int b, pflag;
{
   int bytes;
   void *addr;
   struct mem *pos, *new;

#ifdef DEBUG
   printf("Allocate( %d )\n", b);
#endif
   assert( b<memory_limit-1000 );

   /* round up bytes to multiple of sizeof(struct mem) */
   if (b<MEMSIZ)
     bytes = MEMSIZ;
   else
     bytes = ( (b+MEMSIZ-1) / MEMSIZ ) * MEMSIZ;

   /* try a guess */
   if (guess) {
#ifdef DEBUG
      assert( guess->magic==MAGIC );
      assert( guess->free==1);
#endif
      if (guess->free && guess->size >= bytes + MEMSIZ) {
	 /* a good guess! */
	 pos = guess;
	 /* found larger block, split into 2 parts, allocating first */
	 new = (struct mem *) ( (char *) pos + MEMSIZ + bytes );
	 /* init new fragment node */
	 new->size = pos->size - bytes - MEMSIZ;
	 new->prev = pos;
	 new->next = pos->next;
	 new->free = 1;
	 new->magic = MAGIC;
	 /* tail pointer */
	 if (pos->next)
	   pos->next->prev = new;
	 else
	   tail = new;
	 /* links to pos */
	 pos->next = new;
	 pos->size = bytes;
	 pos->free = 0;
	 memory_used += bytes + MEMSIZ;
	 /* reset guess for next time */
	 guess = new;
	 
	 if (pflag) {
	    /* remove mem node from linked list */
	    remove_mem( pos );
	 }
#ifdef DEBUG
	 check_memory();
#endif
	 return (void *) (pos+1);
      }
   }
   
   /* find first fit */
   pos = head;
   do {
      if (pos->free && pos->size == bytes) {
	 /* found exact fit! */
	 pos->free = 0;
	 memory_used += bytes;
	 if (pflag) {
	    /* remove mem node from linked list */
	    remove_mem( pos );
	 }
	 if (guess==pos)
	   guess = NULL;
#ifdef DEBUG
	 printf("exact fit 0x%x 0x%x\n", pos, pos+1);
	 check_memory();
#endif
	 return (void *) (pos+1);
      }
      else if (pos->free && pos->size >= bytes + MEMSIZ) {
	 /* found larger block, split into 2 parts, allocating first */
	 new = (struct mem *) ( (char *) pos + MEMSIZ + bytes );
	 /* init new fragment node */
	 new->size = pos->size - bytes - MEMSIZ;
	 new->prev = pos;
	 new->next = pos->next;
	 new->free = 1;
	 new->magic = MAGIC;
	 /* tail pointer */
	 if (pos->next)
	   pos->next->prev = new;
	 else
	   tail = new;
	 /* links to pos */
	 pos->next = new;
	 pos->size = bytes;
	 pos->free = 0;
	 memory_used += bytes + MEMSIZ;
	 /* reset guess */
	 if (!guess) {
	    guess = new;
	 }
	 if (pflag) {
	    /* remove mem node from linked list */
	    remove_mem( pos );
	 }
#ifdef DEBUG
	 printf("big fit 0x%x 0x%x\n", pos, pos+1);
	 check_memory();
#endif
	 return (void *) (pos+1);
      }
      else {
	 pos = pos->next;
      }
   } while (pos);

   /* couldn't find block large enough, return NULL */
   guess = NULL;
   return NULL;

}




static void dealloc( addr, b )
void *addr;
int b;
{
   int bytes;
   struct mem *pos, *pred, *succ;

#ifdef DEBUG
   printf("Deallocate( 0x%x, %d )\n", addr, b );
#endif
   assert( addr!=NULL );
   pos = (struct mem *) ( (char *) addr - MEMSIZ );

#ifdef DEBUG
   /* Sanity Checks: */
   assert( pos->magic==MAGIC );
   assert( pos->free==0 );
#endif

   if (b<0) {
      bytes = pos->size;
   }
   else if (b<MEMSIZ) {
      bytes = MEMSIZ;
   }
   else {
      /* round up bytes to multiple of sizeof(struct mem) */
      bytes = ( (b+MEMSIZ-1) / MEMSIZ ) * MEMSIZ;
      assert( pos->size==bytes );
   }

   /* mark as free */
   pos->free = 1;
   memory_used -= bytes;

   /* try to merge this block with successor */
   if (pos->next && pos->next->free==1) {
#ifdef DEBUG
      printf("Merge with successor\n");
#endif
      succ = pos->next;
      pos->size += MEMSIZ + succ->size;
      pos->next = succ->next;
      pos->free = 1;
      if (succ->next)
	 succ->next->prev = pos;
      else
	 tail = pos;
      /* update guess if necessary */
      if (succ==guess) {
	 guess = pos;
      }
      memory_used -= MEMSIZ;
   }

   /* try to merge this block with predecessor */
   if (pos->prev && pos->prev->free==1) {
#ifdef DEBUG
      printf("Merge with predecessor\n");
#endif
      pred = pos->prev;
      pred->size += MEMSIZ + pos->size;
      pred->next = pos->next;
      if (pos->next)
	 pos->next->prev = pred;
      else
	 tail = pred;
      /* update guess if necessary */
      if (pos==guess) {
	 guess = pred;
      }
      /* update pos */
      pos = pred;
      memory_used -= MEMSIZ;
   }

   /* reset guess */
   guess = pos;
#ifdef DEBUG
   check_memory();
#endif
}



/* To keep track of most recently used graphics: */
static int RecentIsosurf[MAXPARMS];
static int RecentHSlice[MAXPARMS];
static int RecentVSlice[MAXPARMS];
static int RecentCHSlice[MAXPARMS];
static int RecentCVSlice[MAXPARMS];
static int RecentHWind[WINDSLICES];
static int RecentVWind[WINDSLICES];
static int RecentTraj[TRAJGROUPS];

static int AccessTime = 1;   /* access time pseudo-clock */




/*** recent ***********************************************************
   Inform the memory manager of the most recently used graphic.
   Input:  ig - the graphic type: ISOSURF, HSLICE, VWIND, TRAJ, etc.
           ip - which parameter, wind slice, traj group, etc.
**********************************************************************/
void recent( ig, ip )
int ig, ip;
{
   if (ig==ISOSURF)
     RecentIsosurf[ip] = AccessTime;
   else if (ig==HSLICE)
     RecentHSlice[ip] = AccessTime;
   else if (ig==VSLICE)
     RecentVSlice[ip] = AccessTime;
   else if (ig==CHSLICE)
     RecentCHSlice[ip] = AccessTime;
   else if (ig==CVSLICE)
     RecentCVSlice[ip] = AccessTime;
   else if (ig==HWIND)
     RecentHWind[ip] = AccessTime;
   else if (ig==VWIND)
     RecentVWind[ip] = AccessTime;
   else if (ig==TRAJ)
     RecentTraj[ip] = AccessTime;
   else
     printf("error in recent( %d, %d )\n", ig, ip );
   AccessTime++;
}




/*** deallocate_lru ***************************************************
   Find and deallocate the least-recently-used graphic.
   Return:  number of bytes freed or 0 if nothing found to deallocate.
**********************************************************************/
static int deallocate_lru()
{
   int ip, it, t, oldtime, oldig, oldip;
   int bytes, b1, b2, b3;

   oldtime = AccessTime;

   /* find oldest isosurface */
   for (ip=0;ip<NumParms;ip++) {
      t = RecentIsosurf[ip];
      if (t>0 && t<oldtime) {
	 oldtime = t;
	 oldig = ISOSURF;
	 oldip = ip;
      }
   }

   /* find oldest horizontal contour slice */
   for (ip=0;ip<NumParms;ip++) {
      t = RecentHSlice[ip];
      if (t>0 && t<oldtime) {
	 oldtime = t;
	 oldig = HSLICE;
	 oldip = ip;
      }
   }

   /* find oldest vertical contour slice */
   for (ip=0;ip<NumParms;ip++) {
      t = RecentVSlice[ip];
      if (t>0 && t<oldtime) {
	 oldtime = t;
	 oldig = VSLICE;
	 oldip = ip;
      }
   }

   /* find oldest horizontal colored slice */
   for (ip=0;ip<NumParms;ip++) {
      t = RecentCHSlice[ip];
      if (t>0 && t<oldtime) {
	 oldtime = t;
	 oldig = CHSLICE;
	 oldip = ip;
      }
   }

   /* find oldest vertical colored slice */
   for (ip=0;ip<NumParms;ip++) {
      t = RecentCVSlice[ip];
      if (t>0 && t<oldtime) {
	 oldtime = t;
	 oldig = CVSLICE;
	 oldip = ip;
      }
   }

   /* find oldest horizontal wind slice */
   for (ip=0;ip<WINDSLICES;ip++) {
      t = RecentHWind[ip];
      if (t>0 && t<oldtime) {
	 oldtime = t;
	 oldig = HWIND;
	 oldip = ip;
      }
   }


   /* find oldest vertical wind slice */
   for (ip=0;ip<WINDSLICES;ip++) {
      t = RecentVWind[ip];
      if (t>0 && t<oldtime) {
	 oldtime = t;
	 oldig = VWIND;
	 oldip = ip;
      }
   }

   /* find oldest trajectory group */
   for (ip=0;ip<TRAJGROUPS;ip++) {
      t = RecentTraj[ip];
      if (t>0 && t<oldtime) {
	 oldtime = t;
	 oldig = TRAJ;
	 oldip = ip;
      }
   }


   if (oldtime<AccessTime) {
      /* found something to deallocate */
      bytes = 0;

      if (oldig==ISOSURF) {
	 /* deallocate all 'oldip' isosurfaces */
	 for (it=0;it<NumTimes;it++) {
	    bytes += free_isosurface( it, oldip );
	    /*printf("[reclaimed %d] isosurf %d %d\n", bytes,oldip,it);*/
	 }
	 RecentIsosurf[oldip] = 0;
      }
      else if (oldig==HSLICE) {
	 /* dealloc all 'oldip' horizontal contour slices */
	 for (it=0;it<NumTimes;it++) {
	    bytes += free_hslice( it, oldip );
	    /*printf("[reclaimed %d] hslice %d %d\n", bytes,oldip,it);*/
	 }
	 RecentHSlice[oldip] = 0;
      }
      else if (oldig==VSLICE) {
	 /* dealloc all 'oldip' horizontal contour slices */
	 for (it=0;it<NumTimes;it++) {
	    bytes += free_vslice( it, oldip );
	    /*printf("[reclaimed %d] vslice %d %d\n", bytes,oldip,it);*/
	 }
	 RecentVSlice[oldip] = 0;
      }
      else if (oldig==CHSLICE) {
	 for (it=0;it<NumTimes;it++) {
	    bytes += free_chslice( it, oldip );
	    /*printf("[reclaimed %d] chslice %d %d\n", bytes,oldip,it);*/
	 }
	 RecentCHSlice[oldip] = 0;
      }
      else if (oldig==CVSLICE) {
	 for (it=0;it<NumTimes;it++) {
	    bytes += free_cvslice( it, oldip );
	    /*printf("[reclaimed %d] cvslice %d %d\n", bytes,oldip,it);*/
	 }
	 RecentCVSlice[oldip] = 0;
      }
      else if (oldig==HWIND) {
	 for (it=0;it<NumTimes;it++) {
	    bytes += free_hwind( it, oldip );
	    /*printf("[reclaimed %d] hwind %d\n", bytes*3,it);*/
	 }
	 RecentHWind[oldip] = 0;
      }
      else if (oldig==VWIND) {
	 for (it=0;it<NumTimes;it++) {
	    bytes += free_vwind( it, oldip );
	    /*printf("[reclaimed %d] vwind %d\n", bytes*3,it);*/
	 }
	 RecentVWind[oldip] = 0;
      }
      else if (oldig==TRAJ) {
	 del_traj_group( oldip );
	 bytes = 1;
	 RecentTraj[oldip] = 0;
      }
      else {
	 printf("oldig=%d\n", oldig );
	 die("error in deallocate_lru");
      }

      if (bytes==0)
	 bytes = 1;
      return bytes;
   }
   else {
      /*printf("free_lru() failed\n");*/
      return 0;
   }
}




/********************************************/
/***         DEBUGGING FUNCTIONS          ***/
/********************************************/


check_memory()
{
   struct mem *pos, *pred;

   pred = NULL;
   pos = head;
   while (pos) {
      if (pos->free!=1 && pos->free!=0) {
	 printf("bad pos->free %d\n", pos->free);
      }
      if (pos->magic!=MAGIC) {
	 printf("bad magic number in node 0x%x\n", pos );
      }
      if (pos->prev != pred) {
	 printf("bad pred pointer 0x%x should be 0x%x\n", pos->prev, pred );
      }
      if (pos->next==NULL && tail!=pos) {
	 die("bad tail\n");
      }
      if (pred && pred->free==1 && pos->free==1) {
	 die("adjacent free blocks");
      }
      if (pred) {
	 if ((char *) pred + pred->size + MEMSIZ != (char *) pos) {
	    die("Bad size");
	 }
      }

      pred = pos;
      pos = pos->next;
   }

   assert( tail->free==1 );

   if (guess)
     assert( guess->free == 1 );

}



dump_memory()
{
   struct mem *pos;

   pos = head;
   while (pos) {
      printf("node: 0x%x\n", pos );
      printf("  size: %d", pos->size );
      printf("  prev: 0x%x", pos->prev );
      printf("  next: 0x%x", pos->next );
      printf("  free: %d\n", pos->free );
      pos = pos->next;
   }
   printf("tail = 0x%x\n", tail );
   printf("memory used: %d\n", memory_used);
}

